# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" git flow feature commands implementation. """
import sys

import click

from peltak.core import context
from peltak.core import git
from peltak.core import hooks
from peltak.core import log
from peltak.core import shell
from . import common


def start(name: str):
    """ Start working on a new feature by branching off develop.

    This will create a new branch off develop called feature/<name>.

    Args:
        name (str):
            The name of the new feature.
    """
    branch = git.current_branch(refresh=True)
    task_branch = 'task/' + common.to_branch_name(name)

    if branch.type not in ('feature', 'hotfix'):
        log.err("Task branches can only branch off <33>feature<32> or "
                "<33>hotfix<32> branches")
        sys.exit(1)

    hooks.register.call('pre-task-start', name)
    common.git_checkout(task_branch, create=True)
    hooks.register.call('post-task-start', name)


def update():
    """ Update the feature with updates committed to develop.

    This will merge current develop into the current branch.
    """
    branch = git.current_branch(refresh=True)
    base_branch = common.get_base_branch()

    common.assert_branch_type('task')
    common.git_checkout(base_branch)
    common.git_pull(base_branch)
    common.git_checkout(branch.name)
    common.git_merge(branch.name, base_branch)


def rename(name: str):
    """ Give the currently developed feature a new name.

    Args:
        name (str):
            The new name of the current feature. The current branch will be
            renamed to 'feature/<new_name>'.
    """
    common.assert_branch_type('task')
    common.git_branch_rename('task/' + name.strip().replace(' ', '_'))


def finish(fast_forward: bool):
    """ Merge current feature branch into develop. """
    pretend = context.get('pretend', False)

    if not pretend and (git.staged() or git.unstaged()):
        log.err(
            "You have uncommitted changes in your repo!\n"
            "You need to stash them before you merge the hotfix branch"
        )
        sys.exit(1)

    branch = git.current_branch(refresh=True)
    base = common.get_base_branch()

    prompt = "<32>Merge <33>{}<32> into <33>{}<0>?".format(branch.name, base)
    if not click.confirm(shell.fmt(prompt)):
        log.info("Cancelled")
        return

    common.assert_branch_type('task')

    hooks.register.call('pre-task-finish', branch, base)

    # Merge task into it's base feature branch
    common.git_checkout(base)
    common.git_pull(base)
    common.git_merge(base, branch.name, no_ff=not fast_forward)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(base)

    hooks.register.call('post-task-finish', branch, base)


def merged():
    """ Cleanup a remotely merged branch. """
    base_branch = common.get_base_branch()
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('task')

    hooks.register.call('pre-task-merged', branch, base_branch)

    # Pull feature branch with the merged task
    common.git_checkout(base_branch)
    common.git_pull(base_branch)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(base_branch)

    hooks.register.call('post-task-merged', branch, base_branch)



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" pypi related commands implementation. """
import sys
from os.path import join
from typing import Optional

from peltak.core import conf
from peltak.core import fs
from peltak.core import log
from peltak.core import shell
from peltak.core import util


def upload(target: str):
    """ Upload the release to a pypi server.

    TODO: Make sure the git directory is clean before allowing a release.

    Args:
        target (str):
            pypi target as defined in ~/.pypirc
    """
    log.info("Uploading to pypi server <33>{}".format(target))
    with conf.within_proj_dir():
        shell.run('python setup.py sdist register -r "{}"'.format(target))
        shell.run('python setup.py sdist upload -r "{}"'.format(target))


def gen_pypirc(username: Optional[str] = None, password: Optional[str] = None):
    """ Generate ~/.pypirc with the given credentials.

    Useful for CI builds. Can also get credentials through env variables
    ``PYPI_USER`` and ``PYPI_PASS``.

    Args:
        username (str):
            pypi username. If not given it will try to take it from the
            `` PYPI_USER`` env variable.
        password (str):
            pypi password. If not given it will try to take it from the
            `` PYPI_PASS`` env variable.
    """
    path = join(conf.get_env('HOME'), '.pypirc')
    username = username or conf.get_env('PYPI_USER', None)
    password = password or conf.get_env('PYPI_PASS', None)

    if username is None or password is None:
        log.err("You must provide $PYPI_USER and $PYPI_PASS")
        sys.exit(1)

    log.info("Generating <94>{}".format(path))

    fs.write_file(path, util.remove_indent('''
        [distutils]
        index-servers = pypi
        
        [pypi]
        repository: https://upload.pypi.org/legacy/
        username: {username}
        password: {password}
        
    '''.format(
        username=username,
        password=password
    )))



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Implementation. """

import re
import textwrap
from collections import OrderedDict
from typing import Dict, List, Optional, Pattern

from peltak.core import conf
from peltak.core import git
from peltak.core import shell
from peltak.core import util
from peltak.core import versioning
from .types import ChangelogItems, ChangelogTag


DEFAULT_TAGS = (
    {"tag": "feature", "header": "Features"},
    {"tag": "change", "header": "Changes"},
    {"tag": "fix", "header": "Fixes"},
)
DEFAULT_TAG_FORMAT = '({tag})'
DEFAULT_CONTINUATION_TAG = '_more'


@util.mark_experimental
def changelog(
    start_rev: Optional[str] = None,
    end_rev: Optional[str] = None,
    title: Optional[str] = None,
) -> str:
    """ Print changelog for given git revision range. """
    # Skip 'v' prefix
    changelog_items = _get_all_changelog_items(start_rev, end_rev)
    return _render_changelog(title, changelog_items)


def _get_all_changelog_items(
    start_rev: Optional[str],
    end_rev: Optional[str]
) -> ChangelogItems:
    commits = _get_commits_in_range(start_rev, end_rev)
    tags = [ChangelogTag(**x) for x in conf.get("changelog.tags", DEFAULT_TAGS)]
    results: ChangelogItems = OrderedDict((tag.header, []) for tag in tags)

    for commit in commits:
        full_message = f"{commit.title}\n\n{commit.desc}"
        commit_items = extract_changelog_items(full_message, tags)
        for header, items in commit_items.items():
            results[header] += items

    return results


def _get_commits_in_range(
    start_rev: Optional[str],
    end_rev: Optional[str]
) -> List[git.CommitDetails]:
    if not start_rev:
        versions = [x for x in git.tags() if versioning.is_valid(x[1:])]
        start_rev = versions[-1] if versions else ''

    if not end_rev:
        end_rev = 'HEAD'

    cmd = 'git log --format=%H'
    if start_rev and end_rev:
        cmd += f" {start_rev}..{end_rev}"
    elif end_rev:
        cmd += f" {end_rev}"

    hashes = shell.run(cmd, capture=True).stdout.strip().splitlines()
    return [git.CommitDetails.get(h) for h in hashes]


def _render_changelog(title: Optional[str], changelog_items: ChangelogItems) -> str:
    if title is None:
        # title is None if title parameter wasn't pass. To show empty title
        # you can use --title ''
        title = f"v{versioning.current()}"

    lines = []
    if title:
        lines += [
            f"<35>{title}<0>",
            f"<32>{'=' * (len(title) + 1)}<0>",
            "",
        ]

    for header, items in changelog_items.items():
        if items:
            lines += [
                "",
                f"<32>{header}<0>",
                f"<32>{'-' * len(header)}<0>",
                "",
            ]
            for item_text in items:
                # Newlines are only inserted via continuation tags, all other
                # newlines are removed during changelog items extraction.
                blocks = item_text.splitlines()
                item_content = '\n'.join(
                    '\n'.join(textwrap.wrap(b, width=77)) for b in blocks
                )
                item_content = textwrap.indent(item_content, prefix='  ')
                item_lines = item_content.splitlines()
                lines.append(f"- {item_lines[0].lstrip()}")
                lines += item_lines[1:]

            lines += [""]

    return '\n'.join(lines)


def extract_changelog_items(text: str, tags: List[ChangelogTag]) -> Dict[str, List[str]]:
    """ Extract all tagged items from text.

    Args:
        text (str):
            Text to extract the tagged items from. Each tagged item is a
            paragraph that starts with a tag. It can also be a text list item.

    Returns:
        tuple[list[str], list[str], list[str]]:
            A tuple of ``(features, changes, fixes)`` extracted from the given
            text.

    The tagged items are usually features/changes/fixes but it can be configured
    through `pelconf.yaml`.
    """
    tag_format = conf.get("changelog.tag_format", DEFAULT_TAG_FORMAT)
    continuation_tag = conf.get("changelog.continuation_tag", DEFAULT_CONTINUATION_TAG)
    patterns = {
        tag.header: tag_re(tag_format.format(tag=tag.tag))
        for tag in tags
    }
    more_pttrn = tag_re(tag_format.format(tag=continuation_tag))
    items: ChangelogItems = {tag.header: [] for tag in tags}
    curr_tag = None
    curr_text = ''
    last_tag = None

    for line in text.splitlines():
        if not line.strip():
            if curr_tag is not None:
                items[curr_tag].append(curr_text)
                curr_text = ''
            last_tag = curr_tag
            curr_tag = None

        more_match = more_pttrn.match(line)

        if more_match and last_tag:
            # If it's a continuation tag, then just add it's text to the last
            # used tag. This only works if there was a previous tag.
            curr_tag = last_tag
            curr_text = items[last_tag][-1]
            items[last_tag] = items[last_tag][:-1]
            line = more_match.group('text')
        else:
            for tag in tags:
                m = patterns[tag.header].match(line)
                if m:
                    if curr_tag is not None:
                        # If we're already in a tag definition and we encountered
                        # a beginning of a new tag, just finish by adding new
                        # item to the current tag item list.
                        items[curr_tag].append(curr_text)
                        curr_text = ''
                    curr_tag = tag.header
                    line = m.group('text')
                    break

        if curr_tag is not None:
            if more_match:
                curr_text = '{}\n{}'.format(curr_text, line.strip()).strip()
            else:
                curr_text = '{} {}'.format(curr_text.strip(), line.strip()).strip()

    if curr_tag is not None:
        items[curr_tag].append(curr_text)

    return items


def tag_re(tag: str) -> Pattern:
    """ Return a regular expression to match tagged lines.

    Args:
        tag (str):
            A tag for which you need the regex pattern. This should be a single
            word.

    Returns:
        Pattern: Regex patter object as returned by `re.compile`.

    The returned pattern will match only the starting line that contains the
    tag. The item definition might span across multiple lines and this has to
    be handled separately
    """
    return re.compile(
        # r'\(feature\) (?P<text>.*?\n\n)',
        r'(- |\* |\s+)?{tag} (?P<text>.*)'.format(tag=re.escape(tag)),
    )



# Copyright 2017-2020 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
####################
``peltak changelog``
####################

CLI definition.
"""

from peltak.commands import root_cli, click


@root_cli.group(
    'changelog',
    invoke_without_command=True,
    help=(
        "Generate changelog from commit messages.\n"
        "\n"
        "This will go over all commits between the given git revisions and"
        "extract all changelog data from them. If called without arguments "
        "it will show all changelog entries since the last version tag."
    )
)
@click.option(
    '-s', '--start-rev',
    help="Starting revision. Defaults to last version tag found"
)
@click.option(
    '-e', '--end-rev',
    help='End revision, defaults to HEAD',
)
@click.option(
    '-t', '--title',
    default=None,
    help=(
        'Changelog title.\n'
        '\n'
        'Will use the current project version if not given. Use empty string '
        'to remove the title completely, ie `peltak changelog --title ""`'
    ),
)
@click.pass_context
def changelog_cli(ctx: click.Context, start_rev: str, end_rev: str, title: str) -> None:
    """ Generate changelog from commit messages. """
    if ctx.invoked_subcommand:
        return

    from peltak.core import shell
    from . import logic

    changelog = logic.changelog(start_rev=start_rev, end_rev=end_rev, title=title)
    shell.cprint(changelog)
